from pathlib import Path

import emfile

from ..utils.generic import WriteError, listify


def write_em(image, file_path, overwrite=False):
    """
    write an image to disk as an .em file
    """
    image = listify(image)
    file_path = Path(file_path)
    if len(image) != 1:
        raise WriteError("Cannot write multiple images to the same path.")
    image = image[0]

    if not file_path.suffix:
        file_path = file_path.with_suffix(".em")

    # TODO: pass pixel size and other stuff to header
    emfile.write(file_path, image.data, header_params={}, overwrite=overwrite)



import dask.array as da
import tifffile
from cryotypes.image import Image, validate_image

from ..utils.generic import guess_name


def read_tif(image_path, name_regex=None, lazy=True, **kwargs):
    """
    read an tif file
    """
    name = guess_name(image_path, name_regex)

    with tifffile.TiffFile(image_path) as tif:
        if lazy:
            data = da.from_zarr(tif.aszarr())
        else:
            data = tif.asarray()
        pixel_size = (
            10000 / tif.pages.first.get_resolution(tifffile.RESUNIT.MICROMETER)[-1]
        )

    stack = data.ndim == 3

    img = Image(
        data=data,
        experiment_id=name,
        pixel_spacing=pixel_size,
        source=image_path,
        stack=stack,
    )

    return validate_image(img, coerce=True)



import json

import numpy as np
import pandas as pd
from cryotypes.poseset import validate_poseset

from ..utils.generic import guess_name
from ..utils.types import PoseSet


def read_eman2_json(
    json_path,
    name_regex=None,
    center_on_tomo=None,
    **kwargs,
):
    with open(json_path) as f:
        data = json.load(f)

    coords = np.array([el[:3] for el in data["boxes_3d"]], dtype=float)
    if not len(coords):
        coords = np.empty((0, 3), dtype=float)

    method = np.array([el[3] for el in data["boxes_3d"]], dtype=str)
    score = np.array([el[4] for el in data["boxes_3d"]], dtype=float)
    particle_class = np.array([el[5] for el in data["boxes_3d"]], dtype=int)

    features = pd.DataFrame({"method": method, "score": score, "class": particle_class})
    class_features = pd.DataFrame(
        {int(k): v for k, v in data["class_list"].items()}
    ).T.reset_index(names="class")

    features = pd.merge(features, class_features, on="class")

    name = guess_name(json_path, name_regex)
    px_size = data["apix_unbin"]

    if center_on_tomo is not None:
        import h5py

        with h5py.File(center_on_tomo, mode="r") as hdf:
            img = hdf["MDF"]["images"]["0"]
            shape = np.array(img["image"].shape)[::-1]  # invert zyx to xyz
            tomo_px_size = img.attrs["EMAN.apix_x"].item()

        # shift to center
        coords = ((shape * tomo_px_size / 2) + (coords * px_size)) / px_size

    poseset = PoseSet(
        position=coords,
        experiment_id=name,
        pixel_spacing=px_size,
        source=json_path,
        features=features,
    )
    return validate_poseset(poseset, coerce=True)



import mrcfile
import numpy as np
import starfile

from cryohub.utils.testing import assert_dataframe_equal
from cryohub.writing.main import write


def test_write_path(tmp_path, image_stack, volume, poseset, relion40_star):
    mrc_path1 = tmp_path / "test1.mrc"
    mrc_path2 = tmp_path / "test2.mrc"
    star_path = tmp_path / "test.star"

    write(image_stack, mrc_path1)
    write(volume, mrc_path2)
    write(poseset, star_path)

    np.testing.assert_array_equal(mrcfile.open(mrc_path1).data, image_stack.data)
    np.testing.assert_array_equal(mrcfile.open(mrc_path2).data, volume.data)
    for df1, df2 in zip(
        starfile.read(star_path, always_dict=True).values(), relion40_star.values()
    ):
        assert_dataframe_equal(df1, df2)



import h5py
import numpy as np
from cryotypes.image import Image, validate_image

from ..utils.generic import guess_name


def read_eman2_hdf(image_path, name_regex=None, **kwargs):
    """
    read an EMAN2 hdf file
    """
    name = guess_name(image_path, name_regex)

    with h5py.File(image_path, mode="r") as hdf:
        # TODO: how does this vary?
        img = hdf["MDF"]["images"]["0"]
        data = np.asarray(img["image"])

        px_size = img.attrs["EMAN.apix_x"].item()

    img = Image(
        data=data,
        experiment_id=name,
        pixel_spacing=px_size,
        source=image_path,
        stack=False,
    )

    return validate_image(img, coerce=True)



from pathlib import Path

import h5py

from ..utils.generic import WriteError, listify


def write_hdf(image, file_path, overwrite=False):
    """
    write an image to disk as an .hdf file
    """
    image = listify(image)
    file_path = Path(file_path)
    if len(image) != 1:
        raise WriteError("Cannot write multiple images to the same path.")
    image = image[0]

    if not file_path.suffix:
        file_path = file_path.with_suffix(".hdf")

    mode = "w" if overwrite else "x"

    with h5py.File(file_path, mode=mode) as hdf:
        # TODO: how does this vary?
        (
            hdf.create_group("MDF")
            .create_group("images")
            .create_group("0")
            .create_dataset("image", data=image.data)
        )



import click


@click.group(name="cryohub", context_settings=dict(help_option_names=["-h", "--help"]))
@click.version_option()
def cli():
    """
    Cryohub command line interface.
    """
    pass


def _process_extra_kwargs(kwarg_list):
    kwargs_processed = {}
    try:
        for k, v in zip(kwarg_list[::2], kwarg_list[1::2], strict=True):
            if not k.startswith("-"):
                raise click.UsageError(
                    "Only key-value pairs are allowed as extra arguments."
                )

            for typ in (float, int, str):
                try:
                    v = typ(v)
                    break
                except (TypeError, ValueError):
                    continue
                else:
                    raise TypeError('could not convert "{v}" to a usable type')
            kwargs_processed[k.strip("-").replace("-", "_")] = v
    except ValueError:
        raise click.UsageError("Only key-value pairs are allowed as extra arguments.")
    return kwargs_processed


@cli.command(context_settings=dict(ignore_unknown_options=True))
@click.argument("input", type=click.Path(exists=True, dir_okay=True, resolve_path=True))
@click.argument("output", type=click.Path(dir_okay=False, resolve_path=True))
@click.option(
    "-n",
    "--name-regex",
    metavar="regex",
    help=r"a regex used to infer DataBlock names from paths [fallback: \d+]",
)
@click.option(
    "-f", "--overwrite", is_flag=True, help="Overwrite existing output files."
)
@click.argument("kwargs", nargs=-1, type=click.UNPROCESSED)
def convert(input, output, name_regex, overwrite, kwargs):
    """Convert files between available formats."""
    from pathlib import Path

    import cryohub

    output = Path(output)
    if output.exists() and not overwrite:
        raise click.UsageError(f"{output} already exists. Use -f to overwrite.")

    data = cryohub.read(  # noqa: F841
        input,
        name_regex=name_regex,
        **_process_extra_kwargs(kwargs),
    )
    cryohub.write(data, output, overwrite=overwrite)


@cli.command()
@click.argument("paths", nargs=-1)
@click.option(
    "-n",
    "--name-regex",
    metavar="regex",
    help=r"a regex used to infer DataBlock names from paths [fallback: \d+]",
)
@click.option(
    "--strict", is_flag=True, help="immediately fail if a matched path cannot be read"
)
@click.option(
    "--lazy", is_flag=True, default=True, help="read data lazily (if possible)"
)
def view(paths, name_regex, strict, lazy):
    r"""
    Open files and land in an interactive ipython shell.

    Loaded data is stored in the `data` variable.

    PATHS: any number of files [default='.']

    EXAMPLES:

    Open a .star file as particles:

        cryohub particles.star

    Open particles and images from a directory:

        cryohub /dir/with/star_and_mrc_files/

    Match files such as MyProtein_10.star and MyProtein_001.mrc,
    and name the respective data objects Protein_10 and Protein_001:

        cryohub /path/to/dir/MyProtein* -n 'Protein_\d+'
    """
    from inspect import cleandoc

    from IPython.terminal.embed import InteractiveShellEmbed

    import cryohub

    if not paths:
        paths = ["./*"]

    data = cryohub.read(  # noqa: F841
        *paths,
        name_regex=name_regex,
        strict=strict,
        lazy=lazy,
    )

    # set up ipython shell nicely
    banner = """
        === cryohub ===
        - loaded data is in the list `data`
    """
    # sh.instance() needed due to reggression in ipython
    # https://github.com/ipython/ipython/issues/13966#issuecomment-1696137868
    sh = InteractiveShellEmbed.instance(banner2=cleandoc(banner))
    sh()



from pathlib import Path

from cryotypes.image import ImageProtocol
from cryotypes.poseset import PoseSetProtocol

from ..utils.generic import WriteError, listify
from .em import write_em
from .hdf import write_hdf
from .mrc import write_mrc
from .star import write_star
from .tbl import write_tbl

# a mapping of file extensions to writers, tuple map to tuples (don't forget trailing comma!):
#   - multiple extensions values in the keys use the same writers
#   - multiple writers are called in order from highest to lowers in case previous ones fail
#   - empty suffix must be in the last option (fallback)
# TODO: put this directly in the writers to make it plug and play?
writers_poses = {
    (".tbl",): (write_tbl,),
    (".star", ""): (write_star,),
}

writers_images = {
    (".hdf",): (write_hdf,),
    (".em",): (write_em,),
    (".mrc", ".mrcs", ".st", ".map", ""): (write_mrc,),
}


def write(data, file_path, **kwargs):
    """
    write a single file with the appropriate function
    """
    file_path = Path(file_path)
    data = listify(data)
    if all(isinstance(d, PoseSetProtocol) for d in data):
        writers = writers_poses
    elif all(isinstance(d, ImageProtocol) for d in data):
        writers = writers_images
    else:
        raise WriteError(f"No writer available for object of type {data[0].__class__}")

    for ext, funcs in writers.items():
        if file_path.suffix in ext:
            for func in funcs:
                try:
                    func(data, file_path, **kwargs)
                    return
                except WriteError:
                    # this will be raised by individual writers when the file can't be written.
                    # Keep trying until all options are exhausted
                    continue
    raise WriteError(f"could not write {file_path}")
